extern crate imgui;
extern crate imgui_wgpu;
#[macro_use]
extern crate lazy_static;
extern crate tobj;
extern crate winit_24;

use std::f32::consts::PI;
use std::sync::{Arc, Mutex};
#[cfg(not(target_arch = "wasm32"))]
use std::time::{Duration, Instant};

use cgmath::{
    Decomposed, Deg, Euler, InnerSpace, Point3, Quaternion, Rad, Rotation3, SquareMatrix,
};
use futures::executor::block_on;
use futures::task::LocalSpawn;
use gilrs::Event as GilEvent;
use gilrs::{Gamepad, Gilrs};
use imgui::FontSource;
use imgui::__core::convert::TryInto;
use imgui::*;
use imgui_wgpu::{Renderer as ImguiRenderer, RendererConfig as ImguiRendererConfig};
use legion::systems::{SyncResources, UnsafeResources};
use legion::*;
use rapier3d::counters::Timer;
use rapier3d::dynamics::{
    IntegrationParameters, JointSet, RigidBody, RigidBodyBuilder, RigidBodyHandle, RigidBodySet,
};
use rapier3d::geometry::{BroadPhase, ColliderBuilder, ColliderHandle, ColliderSet, NarrowPhase};
use rapier3d::math;
use rapier3d::math::Point;
use rapier3d::na::{Isometry, Isometry3, Vector, Vector3};
use rapier3d::pipeline::PhysicsPipeline;
use wgpu::{BindGroup, Buffer, TextureView};
use wgpu_subscriber;
use winit_24::event::DeviceEvent::MouseMotion;
use winit_24::platform::unix::x11::ffi::Time;
use winit_24::window::Window;
use winit_24::{
    event::{self, WindowEvent},
    event_loop::{ControlFlow, EventLoop},
};

use crate::camera::{Camera, CameraController};
use crate::components::{Collider, ImguiWindow, LoopState, Physics, Position};
use crate::geometry::load_obj;
use crate::imgui_supp::extended_winit_imgui_support;
use crate::imgui_supp::imgui_support::{ImguiContext, ImguiPlatform};
use crate::owned_event::{OwnedEvent, OwnedEventExtension};
use crate::physics::state::PhysicsState;
use std::fs;

mod camera;
mod components;
mod geometry;
mod imgui_supp;
mod light;
mod owned_event;
mod physics;
mod render;

/*

Collision detection
https://nphysics.org/rigid_body_simulations_with_contacts/

Obj file format
http://paulbourke.net/dataformats/obj/

tobj obj loader
https://docs.rs/tobj/2.0.3/tobj/index.html

mesh generator lib, might be useful
https://docs.rs/genmesh/0.6.2/genmesh/

legion ECS
https://github.com/amethyst/legion


ECS
    animation
    render 3d (good!)
    input/io (yep!)
    collision / physics (yep!)
    entities & behaviours (got the entities!)
    scripting!

 */

//log::info!("");

// ImGUI works on more or less a global state. which is MegaLame
static mut CURRENT_UI: Option<imgui::Ui<'static>> = None;
pub unsafe fn current_ui<'a>() -> Option<&'a imgui::Ui<'a>> {
    CURRENT_UI.as_ref()
}

fn main() {
    let mut world = World::default();

    let mut imgui_prepare_schedule = Schedule::builder()
        .add_system(render::system::imgui_prepare_system())
        .build();

    let mut render_schedule = Schedule::builder()
        .add_system(render::system::render_test_system())
        .build();

    let mut update_schedule = Schedule::builder()
        .add_system(physics::system::update_camera_system())
        .add_system(physics::system::run_physics_system())
        .add_system(physics::system::update_models_system())
        // next system here, gamelogic update system?
        .build();

    let mut event_schedule = Schedule::builder()
        .add_system(owned_event::event_dispatch_system())
        .build();

    let event_loop = EventLoop::<OwnedEventExtension>::with_user_event();
    let mut builder = winit_24::window::WindowBuilder::new();
    builder = builder.with_title("MVGE");

    let window = builder.build(&event_loop).unwrap();

    let mut resources = Resources::default();

    // Load up all the resources
    {
        let mut imgui_context = imgui::Context::create();
        let mut platform = extended_winit_imgui_support::WinitPlatform::init(&mut imgui_context);
        platform.attach_window(
            imgui_context.io_mut(),
            &window,
            extended_winit_imgui_support::HiDpiMode::Default,
        );

        // imgui_supp rendering context
        let mut imgui_context = ImguiContext {
            context: imgui_context,
        };
        let mut imgui_platform = ImguiPlatform { platform: platform };
        let font_size = 20.0 as f32;
        imgui_context.context.io_mut().font_global_scale = 1.0 as f32;
        imgui_context
            .context
            .fonts()
            .add_font(&[FontSource::DefaultFontData {
                config: Some(imgui::FontConfig {
                    oversample_h: 1,
                    pixel_snap_h: true,
                    size_pixels: font_size,
                    ..Default::default()
                }),
            }]);
        imgui_context.context.set_ini_filename(None);

        // The renderer
        let mut renderer = render::state::RenderState::init(&window, &mut imgui_context);
        preload_meshes("./resources");
        entity_loading(&mut world, &mut renderer);

        resources.insert(renderer);
        resources.insert(Arc::new(Mutex::new(imgui_context)));
        resources.insert(Arc::new(Mutex::new(imgui_platform)));
        resources.insert(window);

        // Physics
        let (physics_state, physics_pipeline) =
            PhysicsState::build(rapier3d::math::Vector::new(0.0, -9.81, 0.05));
        resources.insert(physics_state);
        resources.insert(physics_pipeline);

        // Loop data
        resources.insert(LoopState {
            delta_time: Default::default(),
            start_time: Instant::now(),
            step_size: 0.01666, // 60hz
        });

        // And our event stack
        resources.insert(Vec::<OwnedEvent<OwnedEventExtension>>::new());
    };

    let event_loop_proxy = event_loop.create_proxy();

    std::thread::spawn(move || {
        let mut gilrs = Gilrs::new().unwrap();
        // Iterate over all connected gamepads
        let mut gamepad: Option<Gamepad> = None;
        for (_id, gamepad_) in gilrs.gamepads() {
            if gamepad_.name() == "PS4" {
                gamepad = Some(gamepad_);
            }
            // println!(
            //     "{} is {:?} {:?}",
            //     gamepad_.name(),
            //     gamepad_.power_info(),
            //     gamepad_.id()
            // );
        }
        let mut active_gamepad = None;

        loop {
            while let Some(GilEvent { id, event, time }) = gilrs.next_event() {
                //println!("{:?} New event from {}: {:?}", time, id, event);
                active_gamepad = Some(id);
                event_loop_proxy
                    .send_event(OwnedEventExtension::GamepadEvent {
                        gil_event: GilEvent { id, event, time },
                    })
                    .ok();
            }

            // // You can also use cached gamepad state
            // if let Some(gamepad) = active_gamepad.map(|id| gilrs.gamepad(id)) {
            //     if gamepad.is_pressed(Button::South) {
            //         println!("Button South is pressed (XBox - A, PS - X)");
            //     }
            // }

            std::thread::sleep(std::time::Duration::from_millis(50));
        }
    });

    let mut elapsed_time: f32 = {
        // deltatime since last frame
        let loop_state = resources.get::<LoopState>().unwrap();
        loop_state.start_time.elapsed()
    }
    .as_secs_f32();

    let mut delta_time: f32 = 0.0;
    let mut accumulator_time: f32 = 0.0;
    let mut current_time: f32 = elapsed_time;

    event_loop.run(move |event, _, control_flow| {
        *control_flow = ControlFlow::Poll;
        match event {
            event::Event::NewEvents(cause) => {}

            // This is the big boy section of the event loop
            // We : dispatch events and clear the queue, query the loops
            //      time data and prep the dt data. Loop the dt locked
            //      conditionally, and run the fps locked renderer
            event::Event::MainEventsCleared => {
                event_schedule.execute(&mut world, &mut resources);
                imgui_prepare_schedule.execute(&mut world, &mut resources);
                resources
                    .get_mut::<Vec<OwnedEvent<OwnedEventExtension>>>()
                    .unwrap()
                    .clear();

                let (step_size, elapsed_time) = {
                    // deltatime since last frame
                    let loop_state = resources.get::<LoopState>().unwrap();
                    (
                        loop_state.step_size,
                        loop_state.start_time.elapsed().as_secs_f32(),
                    )
                };
                delta_time = elapsed_time - current_time;
                current_time = elapsed_time;
                if delta_time > 0.02 {
                    delta_time = 0.02;
                }
                accumulator_time += delta_time;

                while accumulator_time - step_size >= step_size {
                    accumulator_time -= step_size;
                    // ==== DELTA TIME LOCKED ====
                    update_schedule.execute(&mut world, &mut resources);
                }

                // ==== FPS LOCKED ====
                render_schedule.execute(&mut world, &mut resources);
            }
            // Resizing will queue a request_redraw
            event::Event::WindowEvent {
                event: WindowEvent::Resized(size),
                ..
            } => {
                let width = size.width;
                let height = size.height;

                resources
                    .get_mut::<render::state::RenderState>()
                    .unwrap()
                    .resize(width, height);
            }
            event::Event::WindowEvent {
                event: WindowEvent::CloseRequested,
                ..
            } => *control_flow = ControlFlow::Exit,
            event::Event::RedrawRequested(_) => {
                // Call the render system
                //                imgui_prepare_schedule.execute(&mut world, &mut resources);
                //                render_schedule.execute(&mut world, &mut resources);
            }
            _ => {}
        }

        resources
            .get_mut::<Vec<OwnedEvent<OwnedEventExtension>>>()
            .unwrap()
            .push(event.into());
    });
}

pub fn preload_meshes(resources_path: &str) -> (){
    println!("Preloading meshes...");
    let paths = fs::read_dir(resources_path).unwrap();
    for file in paths {
        let file = file.unwrap().file_name();
        let filename = file.to_str().unwrap();
        if filename.ends_with(".obj") {
            println!("Loading {:?} ...", filename);

        }
    }()
    // I guess it's fine to have them loaded to the gpu
    // But I also want to preserve the raw data
}

pub fn load_colliding_mesh_entity(world: &mut World, renderer: &mut render::state::RenderState, mesh_path: &str) {

    let mut static_floor_body = RigidBodyBuilder::new_static()
        .position(Isometry3::new(Vector3::new(0.0, -8.0, 0.0), Vector::y()))
        .build();

    //let pair = import_mesh("./resources/terrain.obj");
    let pair = load_obj(mesh_path);

    let floor_collider = ColliderBuilder::trimesh(
        pair.0
            .iter()
            .map(|v| {
                let position = v.position();
                Point::<f32>::new(position.x, position.y, position.z)
            })
            .collect(),
        pair.1
            .chunks(3)
            .map(|v| [v[0], v[1], v[2]])
            .collect(),
    )
        .build();

    let plane_mesh = renderer.load_mesh_to_buffer(
        "./resources/terrain.obj",
        Some(wgpu::Color {
            r: 1.0,
            g: 0.7,
            b: 0.3,
            a: 1.0,
        }),
    );

    let plane_entity: Entity = world.push((
        Position {
            x: 0.0,
            y: -8.0,
            z: 0.0,
            rot: Euler {
                x: Deg(0.0),
                y: Deg(0.0),
                z: Deg(0.0),
            },
        },
        plane_mesh,
        Physics {
            rigid_body: static_floor_body,
            rigid_body_handle: None,
        },
        Collider {
            collider: floor_collider,
            collider_handle: None,
        },
    ));

}

pub fn entity_loading(world: &mut World, renderer: &mut render::state::RenderState) {
    let monkey_mesh =
        renderer.load_mesh_to_buffer("./resources/monkey.obj", Some(wgpu::Color::GREEN));
    let light_mesh =
        renderer.load_mesh_to_buffer("./resources/light.obj", Some(wgpu::Color::BLACK));
    let ball_mesh =
        renderer.load_mesh_to_buffer("./resources/ball.obj", Some(wgpu::Color::BLUE));


    let camera_ent: Entity = world.push((
        Camera {
            position: Point3 {
                x: 0.0,
                y: 0.0,
                z: 10.0,
            },
            yaw: Rad(-PI),
            pitch: Rad(PI / 2.0),
        },
        CameraController::new(3.0, 1.0),
    ));



    let light_entity: Entity = world.push((
        Position {
            x: 0.0,
            y: 0.0,
            z: 0.0,
            rot: Euler {
                x: Deg(0.0),
                y: Deg(-25.0),
                z: Deg(0.0),
            },
        },
        light_mesh.clone(),
        renderer.create_light(),
    ));

    let light_entity: Entity = world.push((
        cgmath::Point3 {
            x: -5.0 as f32,
            y: 7.0 as f32,
            z: 10.0 as f32,
        },
        light_mesh,
        renderer.create_light(),
    ));

    let offset = cgmath::vec3(2.0, 2.0, 2.0);
    let transform = Decomposed {
        disp: offset.clone(),
        rot: Quaternion::from_axis_angle(offset.normalize(), Deg(50.0)),
        scale: 1.0,
    };

    let monkey_entity: Entity = world.push((
        Position {
            x: 1.0,
            y: 5.0,
            z: 2.0,
            rot: Euler {
                x: Deg(90.0),
                y: Deg(45.0),
                z: Deg(15.0),
            }, //mx: cgmath::Matrix4::from(transform),
        },
        monkey_mesh,
    ));

    let mut dynamic_ball_body = RigidBodyBuilder::new_dynamic()
        .can_sleep(false)
        .mass(1.0)
        //.position(Isometry3::new(Vector3::new(0.0, 10.0, 0.0), Vector::y()))
        .translation(0.0, 35.0, 0.0)
        .build();


    let pair = load_obj("./resources/ball.obj");

    let ball_collider = ColliderBuilder::trimesh(
        pair.0
            .iter()
            .map(|v| {
                let position = v.position();
                Point::<f32>::new(position.x, position.y, position.z)
            })
            .collect(),
        pair.1
            //.iter()
            .chunks(3)
            .map(|v| [v[0], v[1], v[2]])
            .collect(),
    )
        .build();




    let ball_mesh: Entity = world.push((
        Position {
            x: 0.0,
            y: 20.0,
            z: 0.0,
            rot: Euler {
                x: Deg(25.0),
                y: Deg(45.0),
                z: Deg(15.0),
            },
        },
        ball_mesh,
        Physics {
            rigid_body: dynamic_ball_body,
            rigid_body_handle: None,
        },
        Collider {
            collider: ball_collider,
            collider_handle: None,
        },
    ));
}
